#ifndef ZENOH_NET_
#define ZENOH_NET_

#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>


typedef struct zn_locators_t zn_locators_t;

typedef struct zn_properties_t zn_properties_t;

typedef struct zn_publisher_t zn_publisher_t;

typedef struct zn_query_consolidation_t zn_query_consolidation_t;

typedef struct zn_query_t zn_query_t;

typedef struct zn_query_target_t zn_query_target_t;

typedef struct zn_queryable_t zn_queryable_t;

typedef struct zn_reskey_t zn_reskey_t;

typedef struct zn_scout_t zn_scout_t;

typedef struct zn_session_t zn_session_t;

typedef struct zn_subinfo_t zn_subinfo_t;

typedef struct zn_subscriber_t zn_subscriber_t;

/**
 * A string.
 *
 * Members:
 *   const char *val: A pointer to the string.
 *   unsigned int len: The length of the string.
 *
 */
typedef struct zn_string_t {
  const char *val;
  unsigned int len;
} zn_string_t;

/**
 * An array of bytes.
 *
 * Members:
 *   const unsigned char *val: A pointer to the bytes array.
 *   unsigned int len: The length of the bytes array.
 *
 */
typedef struct zn_bytes_t {
  const unsigned char *val;
  unsigned int len;
} zn_bytes_t;

/**
 * A zenoh-net data sample.
 *
 * A sample is the value associated to a given resource at a given point in time.
 *
 * Members:
 *   zn_string_t key: The resource key of this data sample.
 *   zn_bytes_t value: The value of this data sample.
 */
typedef struct zn_sample_t {
  zn_string_t key;
  zn_bytes_t value;
} zn_sample_t;

/**
 * Information on the source of a reply.
 *
 * Members:
 *   unsigned int kind: The kind of source.
 *   zn_bytes_t id: The unique id of the source.
 */
typedef struct zn_source_info_t {
  unsigned int kind;
  zn_bytes_t id;
} zn_source_info_t;

extern const unsigned int ROUTER;

extern const unsigned int PEER;

extern const unsigned int CLIENT;

extern const unsigned int ALL_KINDS;

extern const unsigned int STORAGE;

extern const unsigned int EVAL;

extern const unsigned int ZN_MODE_KEY;

extern const unsigned int ZN_PEER_KEY;

extern const unsigned int ZN_LISTENER_KEY;

extern const unsigned int ZN_USER_KEY;

extern const unsigned int ZN_PASSWORD_KEY;

extern const unsigned int ZN_MULTICAST_SCOUTING_KEY;

extern const unsigned int ZN_MULTICAST_INTERFACE_KEY;

extern const unsigned int ZN_MULTICAST_ADDRESS_KEY;

extern const unsigned int ZN_SCOUTING_TIMEOUT_KEY;

extern const unsigned int ZN_SCOUTING_DELAY_KEY;

extern const unsigned int ZN_ADD_TIMESTAMP_KEY;

extern const unsigned int ZN_LOCAL_ROUTING_KEY;

extern const unsigned int ZN_INFO_PID_KEY;

extern const unsigned int ZN_INFO_PEER_PID_KEY;

extern const unsigned int ZN_INFO_ROUTER_PID_KEY;

/**
 * Close a zenoh-net session.
 *
 * Parameters:
 *     session: A zenoh-net session.
 */
void zn_close(zn_session_t *session);

/**
 * Notifies the zenoh runtime that there won't be any more replies sent for this
 * query.
 */
void zn_close_query(zn_query_t *query);

/**
 * Create a default set of properties for client mode zenoh-net session configuration.
 * If peer is not null, it is added to the configuration as remote peer.
 *
 * Parameters:
 *   peer: An optional peer locator.
 */
zn_properties_t *zn_config_client(char *peer);

/**
 * Create a default set of properties for zenoh-net session configuration.
 */
zn_properties_t *zn_config_default(void);

/**
 * Create an empty set of properties for zenoh-net session configuration.
 */
zn_properties_t *zn_config_empty(void);

/**
 * Create a default set of properties for peer mode zenoh-net session configuration.
 */
zn_properties_t *zn_config_peer(void);

/**
 * Declare a :c:type:`zn_publisher_t` for the given resource key.
 *
 * Written resources that match the given key will only be sent on the network
 * if matching subscribers exist in the system.
 *
 * Parameters:
 *     session: The zenoh-net session.
 *     resource: The resource key to publish.
 *
 * Returns:
 *    The created :c:type:`zn_publisher_t` or null if the declaration failed.
 */
zn_publisher_t *zn_declare_publisher(zn_session_t *session, zn_reskey_t *reskey);

/**
 * Declare a :c:type:`zn_queryable_t` for the given resource key.
 *
 * Parameters:
 *     session: The zenoh-net session.
 *     resource: The resource key the :c:type:`zn_queryable_t` will reply to.
 *     kind: The kind of :c:type:`zn_queryable_t`.
 *     callback: The callback function that will be called each time a matching query is received.
 *     arg: A pointer that will be passed to the **callback** on each call.
 *
 * Returns:
 *    The created :c:type:`zn_queryable_t` or null if the declaration failed.
 */
zn_queryable_t *zn_declare_queryable(zn_session_t *session,
                                     zn_reskey_t *reskey,
                                     unsigned int kind,
                                     void (*callback)(zn_query_t*, const void*),
                                     void *arg);

/**
 * Associate a numerical id with the given resource key.
 *
 * This numerical id will be used on the network to save bandwidth and
 * ease the retrieval of the concerned resource in the routing tables.
 *
 * Parameters:
 *     session: The zenoh-net session.
 *     resource: The resource key to map to a numerical id.
 *
 * Returns:
 *     A numerical id.
 */
unsigned long zn_declare_resource(zn_session_t *session, zn_reskey_t *reskey);

/**
 * Declare a :c:type:`zn_subscriber_t` for the given resource key.
 *
 * Parameters:
 *     session: The zenoh-net session.
 *     resource: The resource key to subscribe.
 *     sub_info: The :c:type:`zn_subinfo_t` to configure the :c:type:`zn_subscriber_t`.
 *     callback: The callback function that will be called each time a data matching the subscribed resource is received.
 *     arg: A pointer that will be passed to the **callback** on each call.
 *
 * Returns:
 *    The created :c:type:`zn_subscriber_t` or null if the declaration failed.
 */
zn_subscriber_t *zn_declare_subscriber(zn_session_t *session,
                                       zn_reskey_t *reskey,
                                       zn_subinfo_t *sub_info,
                                       void (*callback)(const zn_sample_t*, const void*),
                                       void *arg);

/**
 * Get informations about an zenoh-net session.
 *
 * Parameters:
 *     session: A zenoh-net session.
 *
 * Returns:
 *     A set of properties containing informations on the given zenoh-net session.
 */
zn_properties_t *zn_info(zn_session_t *session);

/**
 * Initialise the zenoh runtime logger
 *
 */
void zn_init_logger(void);

/**
 * Open a zenoh-net session
 *
 * Parameters:
 *     config: A set of properties.
 *
 * Returns:
 *     The created zenoh-net session or null if the creation did not succeed.
 */
zn_session_t *zn_open(zn_properties_t *config);

/**
 * Add a property to a set of properties.
 *
 * Parameters:
 *   ps: A pointer to the properties.
 *   id: The id of the property to add.
 *   value: The value of the property to add.
 *
 * Returns:
 *     A pointer to the updated properties.
 */
zn_properties_t *zn_properties_add(zn_properties_t *ps, unsigned long id, const char *value);

/**
 * Free a set of properties.
 *
 * Parameters:
 *   ps: A pointer to the properties.
 */
void zn_properties_free(zn_properties_t *ps);

/**
 * Get the length of the given properties.
 *
 * Parameters:
 *     ps: A pointer to the properties.
 *
 * Returns:
 *     The length of the given properties.
 */
unsigned int zn_properties_len(zn_properties_t *ps);

/**
 * Return a new empty set of properties.
 */
zn_properties_t *zn_properties_make(void);

/**
 * Get the id of the property at a given index in a set of properties.
 *
 * Parameters:
 *     ps: A pointer to the properties.
 *     n: The index of the property.
 *
 * Returns:
 *     The id of the property at index ``n`` in properties ``ps``.
 */
unsigned int zn_property_id(zn_properties_t *ps, unsigned int n);

/**
 * Get the value of the property at a given index in a set of properties.
 *
 * Parameters:
 *     ps: A pointer to the properties.
 *     n: The index of the property.
 *
 * Returns:
 *     The value of the property at index ``n`` in properties ``ps``.
 */
const zn_bytes_t *zn_property_value(zn_properties_t *ps, unsigned int n);

/**
 * Pull data for a pull mode :c:type:`zn_subscriber_t`. The pulled data will be provided
 * by calling the **callback** function provided to the :c:func:`zn_declare_subscriber` function.
 *
 * Parameters:
 *     sub: The :c:type:`zn_subscriber_t` to pull from.
 */
void zn_pull(zn_subscriber_t *sub);

/**
 * Query data from the matching queryables in the system.
 *
 * Parameters:
 *     session: The zenoh-net session.
 *     resource: The resource key to query.
 *     predicate: An indication to matching queryables about the queried data.
 *     target: The kind of queryables that should be target of this query.
 *     consolidation: The kind of consolidation that should be applied on replies.
 *     callback: The callback function that will be called on reception of replies for this query.
 *     arg: A pointer that will be passed to the **callback** on each call.
 */
void zn_query(zn_session_t *session,
              zn_reskey_t *reskey,
              const char *predicate,
              zn_query_target_t *target,
              zn_query_consolidation_t *consolidation,
              void (*callback)(const zn_source_info_t*, const zn_sample_t*, const void*),
              void *arg);

/**
 * Create a default :c:type:`zn_query_consolidation_t`.
 */
zn_query_consolidation_t *zn_query_consolidation_default(void);

/**
 * Create an incremental consolidation :c:type:`zn_query_consolidation_t`.
 * Multiple occurences of the same resource will be filtered by all routers
 * along the replies path.
 */
zn_query_consolidation_t *zn_query_consolidation_incremental(void);

/**
 * Create an last_hop consolidation :c:type:`zn_query_consolidation_t`.
 * Multiple occurences of the same resource will be filtered by the last router
 * along the replies path.
 */
zn_query_consolidation_t *zn_query_consolidation_last_hop(void);

/**
 * Create a no consolidation :c:type:`zn_query_consolidation_t`.
 * Same resource may be received multiple times from different queriable sources.
 */
zn_query_consolidation_t *zn_query_consolidation_none(void);

/**
 * Return the predicate for this query
 */
const zn_string_t *zn_query_predicate(zn_query_t *query);

/**
 * Return the resource name for this query
 */
const zn_string_t *zn_query_res_name(zn_query_t *query);

/**
 * Create a default :c:type:`zn_query_target_t`.
 */
zn_query_target_t *zn_query_target_default(void);

/**
 * Create a resource key from a resource id.
 *
 * Parameters:
 *     id: The resource id.
 *
 * Returns:
 *     Return a new resource key.
 */
zn_reskey_t *zn_rid(unsigned long id);

/**
 * Create a resource key from a resource id and a suffix.
 *
 * Parameters:
 *     id: The resource id.
 *     suffix: The suffix.
 *
 * Returns:
 *     Return a new resource key.
 */
zn_reskey_t *zn_rid_with_suffix(unsigned long id, const char *suffix);

/**
 * Create a resource key from a resource name.
 *
 * Parameters:
 *     id: The resource name.
 *
 * Returns:
 *     Return a new resource key.
 */
zn_reskey_t *zn_rname(const char *name);

/**
 * The scout mask allows to specify what to scout for.
 */
zn_scout_t *zn_scout(unsigned int what, zn_properties_t *config, unsigned long scout_period);

/**
 * Frees the zn_scout_t by releasing its associated memory.
 */
void zn_scout_free(zn_scout_t *s);

/**
 * Get the number of entities scouted  and available as part of
 * the zn_scout_t
 */
unsigned int zn_scout_len(zn_scout_t *si);

/**
 * Get the locator at the given index.
 */
const char *zn_scout_locator_get(zn_locators_t *ls, unsigned int idx);

/**
 * Get the locators for the scouted.
 */
zn_locators_t *zn_scout_locators(zn_scout_t *si, unsigned int idx);

/**
 * Frees the locators
 */
void zn_scout_locators_free(zn_locators_t *ls);

/**
 * Get the number of locators for the scouted entity.
 */
unsigned int zn_scout_locators_len(zn_locators_t *ls);

/**
 * Get the peer-id for the scouted entity at the given index
 */
const unsigned char *zn_scout_peerid(zn_scout_t *si, unsigned int idx);

/**
 * Get the length of the peer-id for the scouted entity at the given index
 */
unsigned int zn_scout_peerid_len(zn_scout_t *si, unsigned int idx);

/**
 * Get the whatami for the scouted entity at the given index
 */
unsigned int zn_scout_whatami(zn_scout_t *si, unsigned int idx);

/**
 * Sends a reply to a query.
 */
void zn_send_reply(zn_query_t *query,
                   const char *key,
                   const unsigned char *payload,
                   unsigned int len);

/**
 * Create a default subscription info.
 */
zn_subinfo_t *zn_subinfo_default(void);

/**
 * Create a pull mode subscription info.
 */
zn_subinfo_t *zn_subinfo_pull(void);

/**
 * Undeclare a :c:type:`zn_publisher_t`.
 *
 * Parameters:
 *     sub: The :c:type:`zn_publisher_t` to undeclare.
 */
void zn_undeclare_publisher(zn_publisher_t *publ);

/**
 * Undeclare a :c:type:`zn_queryable_t`.
 *
 * Parameters:
 *     sub: The :c:type:`zn_queryable_t` to undeclare.
 */
void zn_undeclare_queryable(zn_queryable_t *sub);

/**
 * Undeclare a :c:type:`zn_subscriber_t`.
 *
 * Parameters:
 *     sub: The :c:type:`zn_subscriber_t` to undeclare.
 */
void zn_undeclare_subscriber(zn_subscriber_t *sub);

/**
 * Write data.
 *
 * Parameters:
 *     session: The zenoh-net session.
 *     resource: The resource key to write.
 *     payload: The value to write.
 *     len: The length of the value to write.
 * Returns:
 *     ``0`` in case of success, ``1`` in case of failure.
 */
int zn_write(zn_session_t *session, zn_reskey_t *reskey, const char *payload, unsigned int len);

#endif /* ZENOH_NET_ */
